import sys
import base64
import pickle
import signal
import inspect
import asyncio
import argparse
import importlib
import aiobotocore
from types import FunctionType
from typing import Callable, Union, Coroutine, Optional
from . import utils
from .worker import BatchWorker
from . import __version__

__description__ = "An AWS SQS processer using asyncio/aiobotocore"


class SQSRetry(Exception):
    def __init__(self, *, max_times: int, after: int = None):
        self.max_times = max_times
        self.seconds_later = after

    def delay_seconds(self, retry_times):
        if self.seconds_later is None:
            return 1 + 2 ** retry_times
        return self.seconds_later


class Message(dict):
    @property
    def message_id(self):
        return self["MessageId"]

    @property
    def body(self):
        if not hasattr(self, "_body"):
            Body = self["Body"]
            self._body = pickle.loads(base64.b64decode(Body))
        return self._body

    @property
    def attributes(self):
        return self.get("Attributes", {})


class SQSProcesser:
    service_name = "sqs"
    MAX_RETRY = 10
    sleep_base = 0.3

    def __init__(
        self,
        queue_name: str,
        message_handler: Union[Callable, Coroutine],
        *,
        region_name: Optional[str] = None,
        concurrency: int = 10,
        visibility_timeout: int = 60,
        idle_sleep: int = 0,
        batch_ops: bool = True,
        client_params: Optional[dict] = None,
        delete_worker_params: Optional[dict] = None,
        change_worker_params: Optional[dict] = None
    ):
        self.queue_name = queue_name
        self.concurrency = concurrency
        self.message_handler = message_handler
        self.visibility_timeout = visibility_timeout
        self.idle_sleep = idle_sleep
        self.QueueUrl = None
        self.hooks = {"after_server_stop": set()}
        client_params = client_params or {}
        client_params["region_name"] = region_name
        self.lock = asyncio.Lock()
        self.session = aiobotocore.get_session()
        self.quit_event = asyncio.Event()
        self.semaphore = asyncio.Semaphore(concurrency)
        self.futures = set()
        self.client = self.session.create_client(self.service_name, **client_params)
        if batch_ops:
            delete_worker_params = delete_worker_params or {}
            delete_worker_params.setdefault("name", "SQSDeleteWorker")
            self.delete_worker = BatchWorker(self.delete_batch, **delete_worker_params)
            change_worker_params = change_worker_params or {}
            change_worker_params.setdefault("name", "SQSChangeWorker")
            self.change_worker = BatchWorker(self.change_batch, **change_worker_params)
        else:
            self.delete_worker = None
            self.change_worker = None

    def __repr__(self):
        return "<{}: queue={}, client={}, concurrency={}, working={}>".format(
            self.__class__.__name__,
            self.queue_name,
            self.client,
            self.concurrency,
            len(self.futures),
        )

    async def _get_queue_url(self) -> str:
        if self.QueueUrl is None:
            async with self.lock:
                if self.QueueUrl is None:
                    resp = await self.client.get_queue_url(QueueName=self.queue_name)
                    self.QueueUrl = resp["QueueUrl"]
        return self.QueueUrl

    async def run_forever(self) -> None:
        while not self.quit_event.is_set():
            try:
                await self._fetch_messages()
            except Exception as e:
                try:
                    await utils.handle_exc(e)
                except Exception:
                    pass
                continue
        await self.close()

    async def close(self) -> None:
        if self.futures:
            await asyncio.wait(self.futures)
        if self.change_worker:
            await self.change_worker.stop()
        if self.delete_worker:
            await self.delete_worker.stop()
        await self.client.close()

    async def _fetch_messages(self) -> None:
        job = self.client.receive_message(
            QueueUrl=(await self._get_queue_url()),
            AttributeNames=["ApproximateReceiveCount"],
            MessageAttributeNames=["All"],
            MaxNumberOfMessages=10,
            VisibilityTimeout=self.visibility_timeout,
            WaitTimeSeconds=20,
        )
        response = await utils.cancel_on_event(job, self.quit_event)
        if self.quit_event.is_set():
            return
        if "Messages" not in response and self.idle_sleep > 0:
            await asyncio.sleep(self.idle_sleep)
        if "Messages" in response:
            for message_dict in response["Messages"]:
                await self.semaphore.acquire()
                fut = asyncio.ensure_future(self.handle(Message(message_dict)))
                self.futures.add(fut)
                fut.add_done_callback(self.futures.remove)

    async def handle(self, message) -> None:
        try:
            delete = True
            try:
                result = self.message_handler(message)
                # if asyncio.iscoroutine(result):
                if inspect.isawaitable(result):
                    await result
            except Exception as e:
                if e.__class__.__name__ == SQSRetry.__name__:
                    receive_count = int(
                        message.attributes.get("ApproximateReceiveCount", 1)
                    )
                    if e.max_times > receive_count:
                        seconds = e.delay_seconds(receive_count)
                        await self.change_one(message, seconds)
                        delete = False
                else:
                    await utils.handle_exc(e)
            finally:
                if delete:
                    await self.delete_one(message)
        finally:
            self.semaphore.release()

    async def change_batch(self, messages: list) -> None:
        entries = [
            {
                "Id": str(index),
                "ReceiptHandle": message["ReceiptHandle"],
                "VisibilityTimeout": timeout,
            }
            for index, (message, timeout) in enumerate(messages)
        ]
        queue_url = await self._get_queue_url()
        n = 0
        while n < self.MAX_RETRY:
            if n > 0:
                await asyncio.sleep(self.sleep_base * (2 ** n))
            try:
                resp = await self.client.change_message_visibility_batch(
                    QueueUrl=queue_url, Entries=entries
                )
            except Exception as e:
                await utils.handle_exc(e)
                n += 1
                if n >= self.MAX_RETRY:
                    raise
                continue
            if "Failed" not in resp:
                return
            failed_ids = set(d["Id"] for d in resp["Failed"])
            await utils.info(f"Change failed {n}: {failed_ids} {resp['Failed']}")
            entries = [entry for entry in entries if entry["Id"] in failed_ids]
            n += 1
        else:
            raise Exception("change_message_visibility_batch failed")

    async def change_one(self, message, visibility_timeout: int):
        if self.change_worker:
            await self.change_worker.put((message, visibility_timeout))
            return
        return await self.client.change_message_visibility(
            QueueUrl=(await self._get_queue_url()),
            ReceiptHandle=message["ReceiptHandle"],
            VisibilityTimeout=visibility_timeout,
        )

    async def delete_batch(self, messages: list) -> None:
        entries = [
            {"Id": str(index), "ReceiptHandle": message["ReceiptHandle"]}
            for index, message in enumerate(messages)
        ]
        queue_url = await self._get_queue_url()
        n = 0
        while n < self.MAX_RETRY:
            if n > 0:
                await asyncio.sleep(self.sleep_base * (2 ** n))
            try:
                resp = await self.client.delete_message_batch(
                    QueueUrl=queue_url, Entries=entries
                )
            except Exception as e:
                await utils.handle_exc(e)
                n += 1
                if n >= self.MAX_RETRY:
                    raise
                continue
            if "Failed" not in resp:
                return
            failed_ids = set(d["Id"] for d in resp["Failed"] if not d["SenderFault"])
            await utils.info(f"Delete failed {n}: {failed_ids}, {resp['Failed']}")
            entries = [entry for entry in entries if entry["Id"] in failed_ids]
            n += 1
        else:
            raise Exception("delete_message_batch failed", messages)

    async def delete_one(self, message):
        if self.delete_worker:
            await self.delete_worker.put(message)
            return
        return await self.client.delete_message(
            QueueUrl=(await self._get_queue_url()),
            ReceiptHandle=message["ReceiptHandle"],
        )

    def after_server_stop(self, func: FunctionType) -> None:
        self.hooks["after_server_stop"].add(func)

    def start(self, loop=None):
        loop = loop or asyncio.get_event_loop()
        loop.add_signal_handler(signal.SIGINT, self.quit_event.set)
        loop.add_signal_handler(signal.SIGTERM, self.quit_event.set)
        try:
            loop.run_until_complete(self._get_queue_url())
        except Exception:
            loop.run_until_complete(self.close())
            loop.run_until_complete(loop.shutdown_asyncgens())
            loop.close()
            print(
                f"The queue {self.queue_name} doesn't exists, exiting...",
                file=sys.stderr,
            )
            sys.exit(1)
        try:
            loop.run_until_complete(self.run_forever())
        finally:
            try:
                for func in self.hooks["after_server_stop"]:
                    if asyncio.iscoroutinefunction(func):
                        loop.run_until_complete(func(loop))
                    else:
                        func(loop)
            except Exception as e:
                loop.run_until_complete(utils.handle_exc(e))
            loop.run_until_complete(loop.shutdown_asyncgens())
            loop.close()


def import_function(path: str) -> FunctionType:
    module_path, func_name = path.rsplit(".", 1)
    module = importlib.import_module(module_path)
    return getattr(module, func_name)


async def plain_handler(message) -> None:
    print(message)


async def rpc_handler(message) -> None:
    record = message.body
    func = import_function(record["fpath"])
    func = getattr(func, "_real", func)
    result = func(*record["args"], **record["kw"])
    if inspect.isawaitable(result):
        result = await result
    log = "{0}:{1}({2},{3})={4}".format(
        message.message_id,
        func.__name__,
        ", ".join([f"{arg!r}" for arg in record["args"]]),
        ", ".join([f"{k}={v}" for k, v in record["kw"].items()]),
        result,
    )
    await utils.info(log)


def main():
    epilog = (
        f"version info: cowfish/{__version__} aiobotocore/{aiobotocore.__version__}"
    )
    parser = argparse.ArgumentParser(description=__description__, epilog=epilog)
    parser.add_argument("queue_name")
    parser.add_argument("region")
    parser.add_argument(
        "-c", "--concurrency", type=int, default=20, help="default to 20"
    )
    parser.add_argument(
        "-v", "--visibility-timeout", type=int, default=60, help="default to 60"
    )
    parser.add_argument("--handler", type=str, help="default to rpc_handler")
    parser.add_argument(
        "--no-batch",
        dest="batch",
        action="store_false",
        default=True,
        help="do not use batch ops",
    )
    args = parser.parse_args()
    if args.handler is None:
        handler = rpc_handler
    else:
        handler = import_function(args.handler)
    processer = SQSProcesser(
        args.queue_name,
        args.region,
        handler,
        concurrency=args.concurrency,
        visibility_timeout=args.visibility_timeout,
        batch_ops=args.batch,
    )
    processer.start()


if __name__ == "__main__":
    main()



import time
import json
import asyncio
import aiobotocore
from typing import Callable, Optional
from .worker import BatchWorker
from . import utils


class Firehose:
    service_name = "firehose"
    MAX_RETRY = 10
    sleep_base = 0.3

    def __init__(
        self,
        stream_name: str,
        region_name: Optional[str] = None,
        encode_func: Optional[Callable] = None,
        delimiter: bytes = b"\n",
        *,
        worker_params: Optional[dict] = None,
        client_params: Optional[dict] = None,
        original_api: bool = False,
    ):
        self.session = aiobotocore.get_session()
        self.stream_name = stream_name
        self.encode_func = encode_func or (lambda o: json.dumps(o).encode())
        self.delimiter = delimiter
        client_params = client_params or {}
        client_params["region_name"] = region_name
        self.client = self.session.create_client(self.service_name, **client_params)
        worker_params = worker_params or {}
        worker_params.setdefault("name", "FirehoseWorker")
        batch_func = self.original_batch if original_api else self.write_batch
        self.worker = BatchWorker(batch_func, **worker_params)

    def __repr__(self):
        return (
            f"<{self.__class__.__name__}: "
            f"stream={self.stream_name}, worker={self.worker!r}>"
        )

    async def put(self, obj) -> None:
        return await self.worker.put(obj)

    async def stop(self) -> None:
        timestamp = time.time()
        await self.worker.stop()
        await self.client.close()
        cost = time.time() - timestamp
        await utils.info(f"{self!r} stopped in {cost:.1f} seconds")

    def _encode(self, obj_list: list) -> bytes:
        encoded = [self.encode_func(obj) for obj in obj_list]
        encoded.append(b"")
        return self.delimiter.join(encoded)

    async def write_batch(self, obj_list: list) -> None:
        record = {"Data": self._encode(obj_list)}
        n = 0
        while n < self.MAX_RETRY:
            if n > 0:
                await asyncio.sleep(self.sleep_base * (2 ** n))
            try:
                await self.client.put_record(
                    DeliveryStreamName=self.stream_name, Record=record
                )
            except Exception as e:
                await utils.handle_exc(e)
                n += 1
                if n >= self.MAX_RETRY:
                    raise
                continue
            return

    async def original_batch(self, obj_list: list) -> None:
        records = [{"Data": self.encode_func(obj) + self.delimiter} for obj in obj_list]
        n = 0
        while n < self.MAX_RETRY:
            if n > 0:
                await asyncio.sleep(self.sleep_base * (2 ** n))
            try:
                resp = await self.client.put_record_batch(
                    DeliveryStreamName=self.stream_name, Records=records
                )
            except Exception as e:
                await utils.handle_exc(e)
                n += 1
                if n >= self.MAX_RETRY:
                    raise
                continue
            if resp["FailedPutCount"] > 0:
                records = [
                    records[i]
                    for i, record in enumerate(resp["RequestResponses"])
                    if "ErrorCode" in record
                ]
                n += 1
                continue
            return
        else:
            raise Exception("write_batch error: firehose put_record_batch failed")



import time
import base64
import pickle
import asyncio
import functools
import aiobotocore
from types import FunctionType
from typing import Callable, Optional, Any, Union
from .worker import BatchWorker
from . import utils


class SQSWriter:
    service_name = "sqs"
    MAX_RETRY = 10
    sleep_base = 0.3

    def __init__(
        self,
        queue_name: str,
        region_name: Optional[str] = None,
        encode_func: Optional[Callable] = None,
        *,
        worker_params: Optional[dict] = None,
        client_params: Optional[dict] = None,
    ):
        self.session = aiobotocore.get_session()
        self.queue_name = queue_name
        self.is_fifo = queue_name.endswith(".fifo")
        self.encode_func = encode_func or (
            lambda o: base64.b64encode(pickle.dumps(o, 2)).decode("ascii")
        )
        client_params = client_params or {}
        client_params["region_name"] = region_name
        self.client = self.session.create_client(self.service_name, **client_params)
        self.QueueUrl = None
        self.lock = asyncio.Lock()
        worker_params = worker_params or {}
        worker_params.setdefault("name", "SQSWorker")
        self.worker = BatchWorker(self.write_batch, **worker_params)

    def __repr__(self):
        return (
            f"<{self.__class__.__name__}: "
            f"queue={self.queue_name}, worker={self.worker!r}>"
        )

    async def _get_queue_url(self) -> str:
        if self.QueueUrl is None:
            async with self.lock:
                if self.QueueUrl is None:
                    resp = await self.client.get_queue_url(QueueName=self.queue_name)
                    self.QueueUrl = resp["QueueUrl"]
        return self.QueueUrl

    async def stop(self) -> None:
        timestamp = time.time()
        await self.worker.stop()
        await self.client.close()
        cost = time.time() - timestamp
        await utils.info(f"{self!r} stopped in {cost:.1f} seconds")

    def _encode(self, obj: Any) -> str:
        return self.encode_func(obj)

    async def write_one(
        self,
        record: Any,
        delay_seconds: int = 0,
        deduplication_id: Optional[str] = None,
        group_id: Optional[str] = None,
        queued: bool = False,
        **attributes,
    ):
        message = {
            "record": record,
            "delay_seconds": delay_seconds,
            "attributes": attributes,
            "params": {},
        }
        if self.is_fifo:
            if deduplication_id:
                message["params"]["MessageDeduplicationId"] = deduplication_id
            if group_id:
                message["params"]["MessageGroupId"] = group_id
        if queued:
            await self.worker.put(message)
            return
        queue_url = await self._get_queue_url()
        return await self.client.send_message(
            QueueUrl=queue_url,
            MessageBody=self._encode(record),
            DelaySeconds=delay_seconds,
            MessageAttributes=attributes,
            **message["params"],
        )

    async def write_batch(self, obj_list: list) -> None:
        Entries = [
            {
                "Id": str(i),
                "MessageBody": self._encode(message["record"]),
                "DelaySeconds": message["delay_seconds"],
                "MessageAttributes": message["attributes"],
                **message["params"],
            }
            for i, message in enumerate(obj_list)
        ]
        queue_url = await self._get_queue_url()
        n = 0
        while n < self.MAX_RETRY:
            if n > 0:
                await asyncio.sleep(self.sleep_base * (2 ** n))
            try:
                resp = await self.client.send_message_batch(
                    QueueUrl=queue_url, Entries=Entries
                )
            except Exception as e:
                await utils.handle_exc(e)
                n += 1
                if n >= self.MAX_RETRY:
                    raise
                continue
            if "Failed" not in resp:
                return
            failed_ids = set(d["Id"] for d in resp["Failed"] if not d["SenderFault"])
            await utils.info(f"Send failed {n}: {failed_ids}, {resp['Failed']}")
            Entries = [entry for entry in Entries if entry["Id"] in failed_ids]
            n += 1
        else:
            raise Exception("write_batch error: SQS send_message_batch failed")

    def async_rpc(
        self,
        func: Union[FunctionType, str, None] = None,
        *,
        delay_seconds: int = 0,
        deduplication_id: Optional[str] = None,
        group_id: Optional[str] = None,
        queued: bool = True,
        **attributes,
    ) -> FunctionType:
        if func is None:
            return functools.partial(
                self.async_rpc,
                delay_seconds=delay_seconds,
                deduplication_id=deduplication_id,
                group_id=group_id,
                queued=queued,
                **attributes,
            )
        if type(func) == str:

            async def async_func(*args, **kw):
                record = {"fpath": func, "args": args, "kw": kw}
                await self.write_one(
                    record,
                    delay_seconds=delay_seconds,
                    deduplication_id=deduplication_id,
                    group_id=group_id,
                    queued=queued,
                    **attributes,
                )

        else:

            @functools.wraps(func)
            async def async_func(*args, **kw):
                fpath = func.__module__ + "." + func.__name__
                record = {"fpath": fpath, "args": args, "kw": kw}
                await self.write_one(
                    record,
                    delay_seconds=delay_seconds,
                    deduplication_id=deduplication_id,
                    group_id=group_id,
                    queued=queued,
                    **attributes,
                )

        return async_func


class StringAttribute(dict):
    def __init__(self, value: str):
        self["DataType"] = "String"
        self["StringValue"] = str(value)


class BinaryAttribute(dict):
    def __init__(self, value: bytes):
        self["DataType"] = "Binary"
        self["BinaryValue"] = bytes(value)


class NumberAttribute(dict):
    def __init__(self, value: str):
        self["DataType"] = "Number"
        self["StringValue"] = str(value)



import os
import time
import json
import base64
import asyncio
import aiobotocore
from typing import Callable, Optional
from .worker import BatchWorker
from . import utils


class Kinesis:
    service_name = "kinesis"
    MAX_RETRY = 10
    sleep_base = 0.3

    def __init__(
        self,
        stream_name: str,
        region_name: Optional[str] = None,
        encode_func: Optional[Callable] = None,
        key_func: Optional[Callable] = None,
        *,
        worker_params: Optional[dict] = None,
        client_params: Optional[dict] = None,
    ):
        self.session = aiobotocore.get_session()
        self.stream_name = stream_name
        self.encode_func = encode_func or (lambda o: json.dumps(o).encode())
        self.key_func = key_func
        client_params = client_params or {}
        client_params["region_name"] = region_name
        self.client = self.session.create_client(self.service_name, **client_params)
        worker_params = worker_params or {}
        worker_params.setdefault("name", "KinesisWorker")
        self.worker = BatchWorker(self.write_batch, **worker_params)

    def __repr__(self):
        return (
            f"<{self.__class__.__name__}: "
            f"stream={self.stream_name}, worker={self.worker!r}>"
        )

    async def stop(self) -> None:
        timestamp = time.time()
        await self.worker.stop()
        await self.client.close()
        cost = time.time() - timestamp
        await utils.info(f"{self!r} stopped in {cost:.1f} seconds")

    def _get_key(self, obj) -> bytes:
        if self.key_func is None:
            return base64.b64encode(os.urandom(24)).decode("ascii")
        return self.key_func(obj)

    def _encode(self, obj) -> bytes:
        return self.encode_func(obj)

    async def write_batch(self, obj_list: list) -> None:
        records = [
            {"PartitionKey": self._get_key(obj), "Data": self._encode(obj)}
            for obj in obj_list
        ]
        n = 0
        while n < self.MAX_RETRY:
            if n > 0:
                await asyncio.sleep(self.sleep_base * (2 ** n))
            try:
                resp = await self.client.put_records(
                    StreamName=self.stream_name, Records=records
                )
            except Exception as e:
                await utils.handle_exc(e)
                n += 1
                if n >= self.MAX_RETRY:
                    raise
                continue
            if resp["FailedRecordCount"] == 0:
                return
            records = [
                records[i]
                for i, record in enumerate(resp["Records"])
                if "ErrorCode" in record
            ]
            n += 1
        raise Exception("write_batch error: kinesis put_records failed")

    async def write_one(self, obj, queued: bool = False):
        if queued:
            await self.worker.put(obj)
            return
        return await self.client.put_record(
            StreamName=self.stream_name,
            Data=self._encode(obj),
            PartitionKey=self._get_key(obj),
        )


class CompactKinesis(Kinesis):
    def __init__(self, *args, **kw):
        import msgpack

        super().__init__(*args, **kw)
        self.buffer = bytearray()
        self.packer = msgpack.Packer()
        self.bufmax = 1024 * 25

    def _encode(self, obj) -> bytes:
        return obj

    async def write(self, obj, flush: bool = False) -> None:
        packed = self.packer.pack(obj)
        if len(self.buffer) + len(packed) >= self.bufmax:
            payload = bytes(self.buffer)
            self.buffer.clear()
            self.buffer.extend(packed)
            await self.write_one(payload, queued=True)
        else:
            self.buffer.extend(packed)
            if flush and self.buffer:
                payload = bytes(self.buffer)
                self.buffer.clear()
                await self.write_one(payload, queued=True)

    async def flush(self) -> None:
        if self.buffer:
            payload = bytes(self.buffer)
            self.buffer.clear()
            await self.write_one(payload, queued=True)

    async def write_fluent(self, label: str, data: str) -> None:
        timestamp = int(time.time())
        packet = (label, timestamp, data)
        await self.write(packet)

    async def stop(self) -> None:
        if self.buffer:
            payload = bytes(self.buffer)
            self.buffer.clear()
            await self.write_one(payload, queued=True)
        await super().stop()
